from pathlib import Path
import json
import typing
from pandas import DataFrame

M_TO_FT = 3.281

with open(Path(__file__).parent / "data/scenarios.json") as f:
    ALL_BUILTIN_SCENARIOS = json.load(f)

ALL_KEYS = [key_ for key_ in ALL_BUILTIN_SCENARIOS.keys()]
ALL_ISSUERS = [value_["issuer"] for _, value_ in ALL_BUILTIN_SCENARIOS.items()]


def _show_builtin_scenarios(format: str = "list") -> typing.Union[list, DataFrame]:
    if format not in ["list", "dataframe"]:
        raise ValueError("The format arg must be either 'list' or 'dataframe'.")
    all_locations = [
        elem_["location name"] for _, elem_ in ALL_BUILTIN_SCENARIOS.items()
    ]
    if format == "list":
        return all_locations
    elif format == "dataframe":
        return (
            # Build a clean dataframe showing what's available as custom scenarios
            DataFrame.from_dict(
                data={
                    "Key": ALL_KEYS,
                    "Location(s) covered": all_locations,
                    "Issuer": ALL_ISSUERS
                }
            )
        )


# Provide list of all locations available in ALL_BUILTIN_SCENARIOS
ALL_LOCATIONS = _show_builtin_scenarios(format="list")


# Check that units are valid
def _check_units(units: str) -> None:
    """Validates units and verifies that unit string descriptor is in the standard set

    Parameters
    ----------
    units : str
        String representing units, can only be one of 'ft', 'in', 'm', and 'cm'

    """
    if not (units in ["ft", "in", "m", "mm", "cm"]):
        raise ValueError(
            f"Units {units} are not supported; only use 'ft', 'in', 'm',"
            f" 'mm', and 'cm'."
        )


# Check that location is valid
def _validate_key(key: typing.Union[str, int]) -> str:
    """Validates location, station, or key given to locate a SLRProjections item

    Parameters
    ----------
    key : typing.Union[str, int]
        A unique identifier referencing one of the builtin scenarios. It can be given as
        either:

        * a location (e.g., 'New Jersey')
        * an int (e.g., '0')
        * a key from the scenarios.json file, e.g., 'cocat-2018-9414290'

        In case multiple matches are possible, the first match will be returned.

    Returns
    -------
    str
        Unique key of the requested scenario

    Examples
    -------
    Use a location:
    >>> utils._validate_key(key='New Jersey')
    'nj-dep-2021'
    Use a key:
    >>> utils._validate_key(key='nj-dep-2021')
    'nj-dep-2021'
    Use an index:
    >>> utils._validate_key(location=0)
    'nj-dep-2021'
    """
    if isinstance(key, int):
        if not (key in list(range(0, len(ALL_LOCATIONS)))):
            raise IndexError(
                "Index notation exceeds length of builtin items available."
            )
        else:
            target_key = ALL_KEYS[key]
    elif isinstance(key, str):
        # Try these:
        if key in ALL_LOCATIONS:
            target_key = ALL_KEYS[ALL_LOCATIONS.index(key)]
        elif key in ALL_KEYS:
            target_key = key
        else:
            raise KeyError(
                "Make sure location is specified either as an "
                "station ID, a key, or a location name."
            )

    return target_key



import numpy as np
import typing
from .utils import _check_units
from .utils import M_TO_FT


# Data class contains the actual projection
class Data:
    def __init__(self, units: str, data: dict) -> None:
        """Data

        Parameters
        ----------
        units : str
            One of the allowable strings that define standard units for SLR
        data : dict
            A Data object that contains 'x' and 'y' keys with 'x' given as years and 'y'
            containing the SLR values for each year

        """
        # Check for length of data
        if len(data["x"]) != len(data["y"]):
            raise ValueError(
                "The two arrays passed in the data dictionary have discordant lengths!"
            )

        # Check the units using the helper function
        _check_units(units=units)

        # Check that the dictionary has the right keys
        if not (("x" in data.keys()) and ("y" in data.keys())):
            raise ValueError("Need 'x' and 'y' keys in the 'data' object")

        # Actually load the data; any null values are converted to nan by imposing dtype
        self.x = np.array(data["x"], dtype=np.float)
        self.y = np.array(data["y"], dtype=np.float)
        self._units = units

    @property
    def units(self):
        return self._units

    def convert(
        self, to_units: str, inplace: bool = False
    ) -> typing.Union[None, np.ndarray]:
        """Convert units in data.y array and will also adjust the units property
        in the Data instance.

        Parameters
        ----------
        to_units : str
            Name of the units to convert toward, must be one of 'm', 'ft', 'in',
            and 'cm'
        inplace : bool, optional
            If true, the values in the data object are overwritten, by default False

        Returns
        -------
        typing.Union[None, np.ndarray]
            Returns either nothing (inplace=True) or an nd.ndarray with the converted
            values

        """
        # Check the units
        _check_units(to_units)

        # Change the units of the y value on the spot
        if self.units == "ft":
            if to_units == "in":
                fac = 12.0
            if to_units == "ft":
                fac = 1.0
            if to_units == "m":
                fac = 1 / M_TO_FT
            if to_units == "cm":
                fac = 1 / M_TO_FT * 100
            if to_units == "mm":
                fac = 1 / M_TO_FT * 1000
        if self.units == "in":
            if to_units == "in":
                fac = 1.0
            if to_units == "ft":
                fac = 1 / 12.0
            if to_units == "m":
                fac = 1 / 12.0 * 1 / M_TO_FT
            if to_units == "cm":
                fac = 1 / 12.0 * 1 / M_TO_FT * 100
            if to_units == "mm":
                fac = 1 / 12.0 * 1 / M_TO_FT * 1000
        if self.units == "m":
            if to_units == "in":
                fac = 12.0 * M_TO_FT
            if to_units == "ft":
                fac = M_TO_FT
            if to_units == "m":
                fac = 1.0
            if to_units == "cm":
                fac = 100.0
            if to_units == "mm":
                fac = 1000.0
        if self.units == "cm":
            if to_units == "in":
                fac = 12.0 * M_TO_FT * 1 / 100.0
            if to_units == "ft":
                fac = M_TO_FT * 1 / 100
            if to_units == "m":
                fac = 1 / 100.0
            if to_units == "cm":
                fac = 1.0
            if to_units == "mm":
                fac = 10.0

        # Apply the transformation
        if inplace:
            self.y = self.y * fac
            self._units = to_units
        else:
            return self.y * fac

    def __repr__(self) -> str:
        s = f"Data in {self.units} ranging from {self.x[0]} to {self.x[-1]}"
        return s



from functools import cache
import json
import urllib.request
import warnings

from numpy import nan
from slr.scenario import Scenario
from pandas import DataFrame

from logging import Logger

logger = Logger(__name__)


class NOAASLR:

    @cache
    def __init__(self, station_id: str = None, **kwargs) -> None:

        # NOAA returns cm if metric is selected; in if english is selected
        _report_year = kwargs.pop("Report Year", 2022)
        _units = kwargs.pop("Data Units", "metric")
        if _units == "metric":
            self.units = "cm"
        else:
            self.units = "in"

        self._station_ID = station_id
        self._base_URL = (
            "https://api.tidesandcurrents.noaa.gov/dpapi/prod/webapi/product"
            "/slr_projections.json"
        )

        # Get the trend object from NOAA API
        link = (
            rf"{self._base_URL}"
            rf"/?station={self._station_ID}"
            rf"&units={_units}"
            rf"&report_year={_report_year}"
        )

        # Load and append
        with urllib.request.urlopen(link) as url:
            try:
                data = json.loads(url.read())["SlrProjections"]
            except ConnectionError:
                warnings.warn("Something came up while retrieving data from NOAA")

        self._data = DataFrame.from_dict(
            data
        )



        # Organize the data in Scenario instance, NOAA provides 4 basic scenarios:
        # * Low
        # * Intermediate-Low
        # * Intermediate
        # * Intermediate-High
        # * High
        # There are confidence intervals but I'm going to skip for now

        noaa_scenarios = {
            "Low" : {
                "Description": "NOAA Low",
                "Short Name": "NOAA Low",
                "Probability" : nan,
            }
        }

        scenario_ = (
            self._data[self._data.loc[:, "scenario"] == "Low"]
        )

        self.sc = Scenario(
            description="NOAA Low",
            short_name="NOAA Low",
            units=self.units,
            probability=.17,
            baseline_year=2005,
            data={
                'x': scenario_.loc[:, "projectionYear"].values,
                'y': scenario_.loc[:, "projectionRsl"].values
            },
        )

    def __repr__(self) -> str:
        return "That's a NOAA SLR object"

if __name__ == "__main__":
    ns = NOAASLR(station_id="9414290")
    print(ns.sc)



import typing

import numpy as np
from pandas import DataFrame, Series

from .data import Data
from .utils import _check_units


# Scenario contains the entire information related to a single trajectory
class Scenario:
    def __init__(
        self,
        description: str,
        short_name: str,
        units: str,
        probability: float,
        baseline_year: int,
        data: dict,
    ) -> None:
        """Scenario represent a specific sea-level rise trajectory, described by the below
        paramaters

        Parameters
        ----------
        description : str
            Long name for the trajectory, e.g., 'High Emission, Low Risk'
        short_name : str
            Short name used for captions, legends, etc. e.g., 'Low Risk'
        units : str
            One of 'm', 'cm', 'in', and 'ft' which describes the units of the 'y' data
        probability : float
            Probability given as the value of the Cumulative Distribution Function (CDF)
            if available, must be between 0 and 1
        baseline_year : int
            The baseline year upon which the sea-level rise trajectory is based.
            Must be an integer
        data : dict
            Dictionary containing the actual sea-level rise data. The dictionary
            must have 'x' and 'y' keys, with values paired as follows:
            * 'x' : [list of years]
            * 'y' : [list of SLR values]
            Both lists must have the same dimension.

        """
        # Units need to be valid
        _check_units(units)

        # Probability is a float or None
        if not (isinstance(probability, float) or (probability is None)):
            raise ValueError(
                "'probability' must be a float or None (null); "
                "check entry in JSON file."
            )

        if isinstance(probability, float):
            if probability >= 0.0 and probability <= 1.0:
                self.probability = probability
            else:
                raise ValueError(
                    f"Probability {self.probability} is not within [0; 1]."
                )

        if probability is None:
            self.probability = np.nan

        # If all checks out, record values
        self.description = description
        self.short_name = short_name
        self.baseline_year = baseline_year
        self.data = Data(units=units, data=data)

    def __repr__(self) -> str:
        s = (
            f"Scenario '{self.short_name}', values are given in {self.units} "
            f"and years range from {self.data.x[0]} to {self.data.x[-1]}"
        )
        return s

    @property
    def units(self):
        return self.data.units

    @property
    def dataframe(self) -> Series:
        """Returns a DataFrame built from x and y in the Scenario

        Returns
        -------
        DataFrame
            DataFrame containing x values as index and y values as values.
        """
        df = DataFrame(
            data={
                f"{self.short_name}, {100. * self.probability:.2f}%"
                f" [{self.units}]": self.data.y
            },
            index=self.data.x,
        )
        df.index.name = f"Year (baseline: {self.baseline_year})"
        return df

    def by_horizon_year(self, horizon_year: typing.Union[int, float]) -> float:
        """Calculates the value of SLR projections by a given horizon_year

        Parameters
        ----------
        horizon_year : int or float
            The value of the year to interpolate the projections

        Returns
        -------
        float
            The interpolated SLR projection at that year, e.g., 2.5.
            Units are implicit and available using Scenario.units

        """
        # Check for horizon year
        if (horizon_year > self.data.x.max()) or (horizon_year < self.data.x.min()):
            raise ValueError(
                "Target year is out of bounds for this location, "
                f"years range from {self.data.x.min()} to "
                f"{self.data.x.max()}."
            )
        # Linearly interpolate value at the horizon_year
        proj = np.interp(horizon_year, self.data.x, self.data.y)
        return proj



from .historical import HistoricalSLR
from .slrprojections import SLRProjections
import matplotlib.pyplot as plt
from typing import Union


def HistoricalVsProjections(
        location: Union[str, int],
        ax: plt.Axes = None
) -> plt.Axes:
    """Will match two objects, HistoricalSLR and SLRProjections
    given a location and make a combined plot. Useful when looking
    for a quick figure.

    Parameters
    ----------
    location : Union[str, int]
        Use the same int, str nomenclature as that used for creating
        a new SLRProjections object, i.e.:
        * an integer, e.g., '0'
        * a station ID, e.g. '9419750'
        * a city e.g., 'Los Angeles, CA'
        * a key from one of the custom scenarios, e.g., nj-dep-2021
    ax : plt.Axes, optional
        _description_, by default None

    Returns
    -------
    plt.Axes
        _description_
    """

    # Create the SLRProjections instance:
    ps = SLRProjections.from_location(location=location)
    # Convert to 'mm'
    ps.convert(to_units='mm', inplace=True)

    # Create the Historical instance:
    hs = HistoricalSLR.from_slrprojections(slrprojections=ps)

    # Create an ax figure if none provided
    if ax is None:
        _, ax = plt.subplots(1, 1)

    ax.plot(
        hs.timeseries.index.year,
        hs.timeseries.values,
        label='Historical trend',
        ls='--', c='k', lw=2
    )
    for scenario_ in ps.scenarios:
        ax.plot(
            scenario_.data.x,
            scenario_.data.y,
            label=scenario_.short_name
        )
    ax.set_xlabel('Year')
    ax.set_ylabel(f'SLR from 2000 in [{ps.units}]')
    ax.set_xlim(
        ax.get_xlim()[0],
        2100
    )
    plt.legend(loc='best')

    return ax



import urllib.request
import json
import warnings
import re
from typing import Union

import numpy as np

from functools import cache
from slr.utils import _check_units
from slr.slrprojections import SLRProjections
from pandas import Timestamp, DataFrame, date_range, DateOffset, Series


class HistoricalSLR:
    """HistoricalSLR will attempt to download and retrieve historical Sea Level
    Rise data for a particular location. Station ID and units must be provided.

    Attributes
    ----------
    station_ID : str, optional
        String describing the NOAA ID, e.g. "9410660", by default None
    units : str, optional
        One of the allowable units to be used for storing the historical trend
        can be either one of 'm', 'mm', 'in', 'cm', or 'ft', by default None

    Methods
    --------
    noaa_properties(format):
        Properties describing the HistoricalSLR object as provided by NOAA.
    from_slrprojections(cls, slrprojections):
        Will attempt to retrieve historical SLR information from NOAA servers using
        a SLRProjections item as seed.s
    """

    @cache
    def __init__(self, station_ID: str = None, units: str = None) -> None:

        """[summary]

        Raises
        ------
        ValueError
            [description]
        """
        _check_units(units)
        self._station_ID = station_ID
        self._base_URL = (
            "https://api.tidesandcurrents.noaa.gov/dpapi/prod/webapi/product"
            "/sealvltrends.json"
        )

        # Get the trend object from NOAA API
        link = rf"{self._base_URL}" rf"/?station={self._station_ID}" rf"&affil=US"

        # Load and append
        with urllib.request.urlopen(link) as url:
            try:
                data = json.loads(url.read())["SeaLvlTrends"][0]
            except ConnectionError:
                warnings.warn("Something came up while retrieving data from NOAA")

        self._data = data

        # Parse data and write to object
        self._zero_year = 2000
        self.trend = data["trend"]
        self.trend_error = data["trendError"]
        self.trend_units = re.findall(r"([a-z]*)[/].", data["units"])[0]
        try:
            self.start_date = Timestamp(data["startDate"])
            self.end_date = Timestamp(data["endDate"])
        except ValueError:
            raise ValueError("Unable to parse start and end dates from response.")

        # Build the historical trend
        self.timeseries = self._build_timeseries()

    def noaa_properties(self, format: str = None) -> Union[DataFrame, str]:
        """Properties describing the HistoricalSLR object as provided by NOAA.

        Parameters
        ----------
        format : str, optional, by default None
            Options for format include:
                - 'dataframe' : will return a pandas DataFrame
                - 'narrative' : will return a formatted string with all properties
                - None : will return properties as dictionary (default)

        Returns
        -------
        Union[DataFrame, str]
            Depending on format option, either:
                - Narrative in plain text format for inclusion in a report, etc.
                - DataFrame with neatly formatted parameters describing the historical
                  trend at the location.

        """
        if format is None:
            return self._data
        elif format == "dataframe":
            df = DataFrame.from_dict(
                data=self._data, orient="index", columns=["Value"]
            ).rename_axis(index="Parameter")
            return df.rename(
                {
                    str_: re.sub(r"(\w)([A-Z])", r"\1 \2", str_).title()
                    for str_ in df.index
                },
                axis=0,
            )
        elif format == "narrative":
            out_str = (
                f"Historical sea-level rise information was retrieved at Station "
                f"{self._station_ID} operated by NOAA CO-OPS. "
                f"The relative sea level trend at that location reads "
                f"{self.trend} {self.trend_units} "
                f"per year with a 95% confidence interval of +/- {self.trend_error} "
                f"{self.trend_units} based on monthly mean sea level data from "
                f"{self.start_date.year} to {self.end_date.year} (approximately "
                f"{self.end_date.year - self.start_date.year} years in total). "
                f"This is equivalent to a change of "
                f"{self.trend * 100} {self.trend_units} in 100 years."
            )
            return out_str

    def _build_timeseries(self) -> Series:

        index = date_range(
            start=Timestamp(self.start_date.year, 6, 15),
            end=Timestamp(self.end_date.year, 6, 15),
            freq=DateOffset(years=1),
        )
        values = np.full_like(index, fill_value=np.nan, dtype=float)
        values[0] = self.trend * (self.start_date.year - self._zero_year)
        values[-1] = self.trend * (self.end_date.year - self._zero_year)

        # Stitch it together
        ts = (
            DataFrame(
                data=values,
                index=index,
                columns=[f"Historical SLR [{self.trend_units}]"],
            )
            .interpolate(method="time")
            .squeeze()
        )
        return ts

    @classmethod
    def from_slrprojections(cls, slrprojections: SLRProjections = None):
        # Read the station ID from the slrprojections
        location = slrprojections.station_ID
        units = slrprojections.units
        if not isinstance(units, str):
            raise TypeError(
                "There are mixed units in the SLRProjections object."
                "Cannot infer units."
            )
        return cls(station_ID=location, units=units)



"""Package configuration."""
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="sealevelrise",
    version="0.1",
    author="Jean O. Toilliez",
    author_email="jeantoilliez@gmail.com",
    description=(
        "Small utility for manipulating Sea Level Rise Projections"
        " for engineering calculations"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.9",
    install_requires=["numpy", "pandas", "matplotlib", "pathlib"],
    extras_require={
        "dev": ["pytest", "pytest-cov", "flake8", "jupyter", "black", "wheel"]
    },
    include_package_data=True,
)



import typing
from copy import deepcopy

from matplotlib.pyplot import Axes, subplots
from pandas import DataFrame, Series, concat

from slr.scenario import Scenario
from slr.utils import (
    ALL_BUILTIN_SCENARIOS,
    _check_units,
    _show_builtin_scenarios,
    _validate_key,
)


# SLRProjections contains SLR Scenario objects for a given location
class SLRProjections:
    def __init__(
        self,
        scenarios: typing.Union[Scenario, typing.List[Scenario]] = None,
        location_name: str = None,
        station_ID: str = None,
        issuer: str = None,
        url: str = None,
        coerce_units: bool = True
    ) -> None:
        """SLRProjections contains SLR scenarios for a specific location defined
        by its name or NOAA ID (preferred).

        Parameters
        ----------
        data : dict, optional
            Data describing the scenarios, by default None
        coerce_units : bool, optional
            If true, will harmonize units in all scenarios provided, by default True

        """

        self.location_name = location_name
        self.station_ID = station_ID
        self.issuer = issuer
        self.url = url
        self.scenarios = scenarios
        self.shape = (len(self.scenarios),)

    @classmethod
    def from_dict(cls, data: dict):
        """Constructs a SLRProjections instance from a dictionary

        Parameters
        ----------
        data : dict
            Dictionary that has the basic info required to build SLRProjections

        Returns
        -------
        SLRProjections
            A new SLRProjections instance
        """

        # Check that you have the right data in there
        if not isinstance(data, dict):
            raise TypeError("data needs to be dictionary")
        if data is not None:
            for attr in ["location name", "station ID (CO-OPS)", "scenarios"]:
                if attr not in data:
                    raise KeyError(f"The {attr} key is missing or mispelled.")
        else:
            raise ValueError("data was passed as None")

        # Record properties; optional are popped
        location_name = data["location name"]
        station_ID = data["station ID (CO-OPS)"]
        issuer = data["issuer"]
        # Optional properties
        url = data.pop('URL', None)

        # Build the scenarios from the dictionary
        scenarios_data = data["scenarios"]

        # Record all Scenario objects into a list of Scenario objects
        scenarios_list = list()

        # Iterate over each items in the dictionary
        for scenario_ in scenarios_data:
            scenarios_list.append(
                Scenario(
                    description=scenario_["description"],
                    short_name=scenario_["short name"],
                    units=scenario_["units"],
                    probability=scenario_["probability (CDF)"],
                    baseline_year=scenario_["baseline year"],
                    data=scenario_["data"],
                )
            )

        return cls(
            scenarios=scenarios_list,
            location_name=location_name,
            station_ID=station_ID,
            issuer=issuer,
            url=url
        )

    @classmethod
    def from_builtin(cls, key: typing.Union[str, int]):
        """Generates a SLRProjections isinstance from one of the builtin scenarios.

        Parameters
        ----------
        key : typing.Union[str, int]
        A unique key defining the builtin scenarios item. It can be given as
        either:

        * a location (e.g., 'New Jersey')
        * an int (e.g., '0')
        * a key from the scenarios.json file, e.g., 'cocat-2018-9414290'

        In case multiple matches are possible, the first match will be returned.

        Returns
        -------
        SLRProjections
            SLRProjections instance corresponding to the key provided
        """
        target_key = _validate_key(key=key)
        return cls.from_dict(data=ALL_BUILTIN_SCENARIOS[target_key])

    @classmethod
    def from_noaa(cls, noaa_id: str = None):
        """Generates a SLRProjections instance from a NOAA ID using NOAA API.
        The primary identifier is the NOAA station ID.

        Parameters
        ----------
        noaa_id : str
            _description_

        Raises
        ------
        NotImplemented
            _description_
        """

        # Invoke dedicated helper function in this one.

        raise NotImplementedError("NOAA connector not implemented")

    @staticmethod
    def show_all_builtin_scenarios(
        format: str = "list",
    ) -> typing.Union[str, DataFrame]:
        """Simple function that lists all builtin scenarios available

        Parameters
        ----------
        format : str
            Format for the list to be returned, either 'list' or 'dataframe',
            default is 'list'.

        Returns
        -------
        list or DataFrame
            List of locations available for manipulation
        """
        return _show_builtin_scenarios(format=format)

    def __repr__(self) -> str:
        s = (
            f"Sea level rise Projections for {self.location_name} "
            f"issued by {self.issuer}; there are "
            f"{self.shape[0]} Scenario(s) available."
        )
        return s

    def __getitem__(self, key: int) -> Scenario:
        if not isinstance(key, int):
            raise TypeError("Wrong type passed; only int is supported")
        if not (key >= 0 and key <= self.shape[0]):
            raise IndexError(
                f"Index is out of range; shape of the instance is {self.shape}"
            )
        # If everything checks out, return the Scenario of interest within the Pack
        return self.scenarios[key]

    @property
    def dataframe(self) -> DataFrame:
        """Builds a pd.DataFrame from all Scenario objects in this instance

        Returns
        -------
        DataFrame
            A single pd.DataFrame containing all Scenario.data.x and
            Scenario.data.y instances
        """
        df = concat([scenario_.dataframe for scenario_ in self.scenarios], axis=1)
        return df

    @property
    def units(self) -> str:
        """Returns the units used in the Scenario.data instances. If more than one
         units are used, returns a list; otherwise, if the same units are used for
         all Scenario.data instances, a single value is returned.

        Returns
        -------
        str
            The name of the units used in the Scenario.data instances,
            either a list or a single value out of 'm', 'ft', 'in', or 'cm'.
        """
        # Returns the units of each Scenario
        units = []
        for scenario in self.scenarios:
            units.append(scenario.units)

        if all(x == units[0] for x in units):
            return units[0]
        else:
            return units

    def by_horizon_year(
        self, horizon_year: float, merge: bool = True, coerce_errors: bool = False
    ) -> typing.Union[Series, DataFrame]:
        """Generate a Series with projected values for SLR
        for a given horizon year for each Scenario. It is a wrapper of the method
        from the Scenario class

        Parameters
        ----------
        horizon_year : float or int
            Value for the horizon year (e.g. 2055)
        merge: bool, optional
            If set to True (default), will returns a the projected value merged with
            the built-in projections as a pandas.DataFrame instance.
            If set to False, will return a pandas.Series instance with the projected
            values for each Scenario
            The difference is primarily cosmetic
        coerce_errors: bool, optional
            If set to True (default), will coerce linear interpolation errors by
            replacing with np.nan; if set to False, will raise errors

        Returns
        -------
        Series
            List of projected values by the horizon year OR
        DataFrame
            Merged DataFrame with built-in projections and the newly calculated
            projections

        Raises
        ------
        ValueError
            If the horizon year exceeds the range provided in the original data.
        """

        proj = dict()
        for scenario in self.scenarios:
            proj[scenario.short_name] = scenario.by_horizon_year(
                horizon_year=horizon_year
            )
        ds = Series(
            data=proj,
            name=f"SLR at {self.location_name} by {horizon_year} [{self.units}]",
        )

        if not merge:
            # Simply return calculated values
            return ds
        else:
            # Deep copy of existing (built-in) projections
            df = self.dataframe.copy(deep=True)
            # Include new values
            df.loc[horizon_year, :] = ds.values
            df.sort_index(inplace=True)
            return df

    def convert(self, to_units: str, inplace: bool = False) -> DataFrame:
        """Provides on the fly or inplace units conversion for all Scenarios
        within a SLRProjections instance.

        Parameters
        ----------
        to_units : str
            The value of the destination units, which can only be one of:
            * 'm'
            * 'cm'
            * 'in'
            * 'ft'
            * 'mm'
        inplace : bool, optional
            Wether the conversion is 'on the fly' (inplace=False),
            or if that conversion is burnt in the current SLRProjections instance
            (inplace=Trie). By default set to False (on-the-fly behavior)

        Returns
        -------
        DataFrame
            Only returns the DataFrame of the resultant converted set of projections.
        """
        # Check the units
        _check_units(to_units)

        # Scan over Scenario objects and change the units
        if inplace:
            for scenario_ in self.scenarios:
                scenario_.data.convert(to_units=to_units, inplace=True)
            return self.dataframe
        else:
            # create a deep copy and output it
            temp = deepcopy(self)
            for scenario_ in temp.scenarios:
                scenario_.data.convert(to_units=to_units, inplace=True)
            return temp.dataframe

    def plot(self, ax: Axes = None, horizon_year: float = None) -> Axes:

        # Handle ax
        if ax is None:
            _, ax = subplots(1, 1, figsize=(5.35, 3.5))

        # Graphics
        for index, series in self.dataframe.iteritems():
            ax.plot(series, label=series.name)

        # Handle horizon year
        if horizon_year is not None:
            ds = self.by_horizon_year(horizon_year=horizon_year, merge=False)
            ax.axvline(x=horizon_year, **{"ls": "--", "lw": 1, "c": "lightgray"})
            for name, val in ds.iteritems():
                ax.scatter(x=horizon_year, y=val)

        # Plot all values
        ax.set_ylabel(f"SLR [{self.units}]")
        ax.set_title(f"SLR for {self.location_name}")
        ax.legend()

        return ax


if __name__ == "__main__":
    sf = SLRProjections.from_builtin(key='NPCC3-new-york-2019')
    print(sf)
    print(sf.scenarios[1].short_name)
    print(sf.url)
    print(sf.scenarios[1].by_horizon_year(horizon_year=2055))
    print(sf.scenarios[1].units)
    print(SLRProjections.show_all_builtin_scenarios(format='dataframe'))
